#ifndef __LINUX_LOCK_H
#define __LINUX_LOCK_H

#ifndef __LINUX_SPINLOCK_H
# error "do not include lock.h directly"
#endif

#ifdef CONFIG_PREEMPT_RT
#include <linux/rt_lock.h>

extern void __call_bad_lock_func(void);
#define __CHECK_LOCK(l)							\
	do {								\
		if (!__builtin_types_compatible_p(typeof(l), lock_t *))	\
			__call_bad_lock_func();				\
	} while (0)

#define lock_t					rtlock_t
#define __LOCK_UNLOCKED(l)			__RT_LOCK_UNLOCKED(l)
#define LOCK_UNLOCKED(l)			RT_LOCK_UNLOCKED(l)
#define DEFINE_LOCK(l)				DEFINE_RTLOCK(l)
#define lock_init(l) do { __CHECK_LOCK(l); rt_spin_lock_init(l); } while (0)
#define lock_is_locked(l) ({ __CHECK_LOCK(l); rt_lock_is_locked(l); })
#define lock_is_contended(l) ({ __CHECK_LOCK(l); 0; })
#define lock_can_lock(l) ({ __CHECK_LOCK(l); !rt_lock_is_locked(&(l)->lock); })
#define assert_lock_locked(l) ({ __CHECK_LOCK(l); BUG_ON(!rt_lock_is_locked(l)); })
#define cond_resched_lock(l) ({ __CHECK_LOCK(l); cond_resched_rt_lock(l); })
#define atomic_dec_and_lock(a, l) ({ __CHECK_LOCK(l); \
					   rt_atomic_dec_and_spin_lock(l, a); })

#define acquire_lock_nested(l, subclass) do { \
		__CHECK_LOCK(l); rt_spin_lock_nested(l, subclass); } while (0)
#define acquire_lock_nest_lock(l, nest_lock) do {\
		__CHECK_LOCK(l); rt_spin_lock_nest_lock(l, nest_lock); } while (0)
#define acquire_lock_irq(l) do { \
		__CHECK_LOCK(l); rt_spin_lock(l); } while (0)
#define acquire_lock_bh(l) do { \
		__CHECK_LOCK(l); rt_spin_lock(l); } while(0)
#define acquire_lock_irqsave_nested(l, flags, subclass)			\
				do {					\
					(void)flags;			\
					__CHECK_LOCK(l);		\
					rt_spin_lock_nested(l, subclass); \
				} while (0)
#define acquire_lock_irqsave(l, flags)				\
				do {				\
					flags = 0;		\
					__CHECK_LOCK(l);	\
					rt_spin_lock(l);	\
				} while (0)

#define release_lock_irq(l) do { __CHECK_LOCK(l); rt_spin_unlock(l); } while (0)
#define release_lock_wait(l) do { __CHECK_LOCK(l); rt_spin_unlock_wait(l); } while (0)
#define release_lock_bh(l) do { __CHECK_LOCK(l); rt_spin_unlock(l); } while (0)
#define release_lock_irqrestore(l, flags)				\
				do {					\
					(void)flags;			\
					__CHECK_LOCK(l);		\
					rt_spin_unlock(l);		\
				} while (0)

#define trylock_bh(l) ({ __CHECK_LOCK(l); rt_spin_trylock(l); })
#define trylock_irq(l) ({ __CHECK_LOCK(l); rt_spin_trylock(l); })
#define trylock_irqsave(l, flags)				\
				({				\
					(void)flags;		\
					__CHECK_LOCK(l);	\
					rt_spin_trylock(l);	\
				})

#define acquire_lock(l) do { __CHECK_LOCK(l); rt_spin_lock(l); } while (0)
#define trylock(l) ({ __CHECK_LOCK(l); rt_spin_trylock(l); })
#define release_lock(l) do { __CHECK_LOCK(l); rt_spin_unlock(l); } while (0)

#else /* CONFIG_PREEMPT_RT */
#define lock_t					spinlock_t
#define __LOCK_UNLOCKED(l)			__SPIN_LOCK_UNLOCKED(l)
#define LOCK_UNLOCKED(l)			SPIN_LOCK_UNLOCKED(l)
#define DEFINE_LOCK(l)				DEFINE_SPINLOCK(l)
#define lock_init(l)				spin_lock_init(l)
#define lock_is_locked(l)			spin_is_locked(l)
#define lock_is_contended(l)			spin_is_contended(l)
#define lock_can_lock(l)			spin_can_lock(l)
#define assert_lock_locked(l)			assert_spin_locked(l)
#define cond_resched_lock(l)			cond_resched_spinlock(l)
#define atomic_dec_and_lock(a, l)		atomic_dec_and_spinlock(a, l)

#define acquire_lock_nested(l, subclass)	spin_lock_nested(l, subclass)
#define acquire_lock_nest_lock(l, nest_lock)	spin_lock_nest_lock(l, nest_lock)
#define acquire_lock_irqsave(l, flags)		spin_lock_irqsave(l, flags)
#define acquire_lock_irq(l)			spin_lock_irq(l)
#define acquire_lock_bh(l)			spin_lock_bh(l)
#define acquire_lock_irqsave_nested(l, flags, subclass)				\
				spin_lock_irqsave_nested(l, flags, subclass)

#define release_lock_irq(l)			spin_unlock_irq(l)
#define release_lock_wait(l)			spin_unlock_wait(l)
#define release_lock_bh(l)			spin_unlock_bh(l)
#define release_lock_irqrestore(l, flags)	spin_unlock_irqrestore(l, flags)

#define trylock_bh(l)				spin_trylock_bh(l)
#define trylock_irq(l)				spin_trylock_irq(l)
#define trylock_irqsave(l, flags) 		spin_trylock_irqsave(l, flags)

/* The following can not be macros, since they are used in structures */
static inline void acquire_lock(lock_t *l)
{
	spin_lock(l);
}

static inline int trylock(lock_t *l)
{
	return spin_trylock(l);
}

static inline void release_lock(lock_t *l)
{
	spin_unlock(l);
}
#endif /* CONFIG_PREEMPT_RT */

#endif /* __LINUX_SPINLOCK_H */
