#ifndef __LINUX_PERCPU_H
#define __LINUX_PERCPU_H

#include <linux/preempt.h>
#include <linux/slab.h> /* For kmalloc() */
#include <linux/smp.h>
#include <linux/cpumask.h>

#include <asm/percpu.h>

#ifdef CONFIG_SMP
#define DEFINE_PER_CPU(type, name)					\
	__attribute__((__section__(".data.percpu")))			\
	PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name

#ifdef MODULE
#define SHARED_ALIGNED_SECTION ".data.percpu"
#else
#define SHARED_ALIGNED_SECTION ".data.percpu.shared_aligned"
#endif

#define DEFINE_PER_CPU_SHARED_ALIGNED(type, name)			\
	__attribute__((__section__(SHARED_ALIGNED_SECTION)))		\
	PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name		\
	____cacheline_aligned_in_smp

#define DEFINE_PER_CPU_PAGE_ALIGNED(type, name)			\
	__attribute__((__section__(".data.percpu.page_aligned")))	\
	PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name
#else
#define DEFINE_PER_CPU(type, name)					\
	PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name

#define DEFINE_PER_CPU_SHARED_ALIGNED(type, name)		      \
	DEFINE_PER_CPU(type, name)

#define DEFINE_PER_CPU_PAGE_ALIGNED(type, name)		      \
	DEFINE_PER_CPU(type, name)
#endif

#define EXPORT_PER_CPU_SYMBOL(var) EXPORT_SYMBOL(per_cpu__##var)
#define EXPORT_PER_CPU_SYMBOL_GPL(var) EXPORT_SYMBOL_GPL(per_cpu__##var)

/* Enough to cover all DEFINE_PER_CPUs in kernel, including modules. */
#ifndef PERCPU_ENOUGH_ROOM
#ifdef CONFIG_MODULES
#define PERCPU_MODULE_RESERVE	8192
#else
#define PERCPU_MODULE_RESERVE	0
#endif

#define PERCPU_ENOUGH_ROOM						\
	(__per_cpu_end - __per_cpu_start + PERCPU_MODULE_RESERVE)
#endif	/* PERCPU_ENOUGH_ROOM */

/*
 * Must be an lvalue. Since @var must be a simple identifier,
 * we force a syntax error here if it isn't.
 */
#define get_cpu_var(var) (*({				\
	extern int simple_identifier_##var(void);	\
	preempt_disable();				\
	&__get_cpu_var(var); }))
#define put_cpu_var(var) preempt_enable()

#ifdef CONFIG_PREEMPT_RT
# define get_cpu_var_nort(var) per_cpu(var, raw_smp_processor_id())
# define put_cpu_var_nort(var) do { (void)per_cpu__##var; } while (0)
# define __get_cpu_var_nort(var) __raw_get_cpu_var(var)
#else
# define get_cpu_var_nort(var) get_cpu_var(var)
# define put_cpu_var_nort(var) put_cpu_var(var)
# define __get_cpu_var_nort(var) __get_cpu_var(var)
#endif

#if defined(CONFIG_TEST_CPU_LOCKS) || defined(CONFIG_PREEMPT_RT)

struct cpu_lock_t {
	lock_t		lock;
	struct task_struct	*owner;
	int 			depth;
	int			bh;
};

#define DEFINE_CPU_LOCK(lock)			\
	DEFINE_PER_CPU(struct cpu_lock_t, lock##__lock)

#define DECLARE_CPU_LOCK(lock)			\
	DECLARE_PER_CPU(struct cpu_lock_t, lock##__lock)

#define EXPORT_CPU_LOCK_SYMBOL(lock)		\
	EXPORT_PER_CPU_SYMBOL(lock##__lock)

#define EXPORT_CPU_LOCK_SYMBOL_GPL(lock)	\
	EXPORT_PER_CPU_SYMBOL_GPL(lock##__lock)

/* PREEMPT_RT should not disable bottom halves */
#ifndef CONFIG_PREEMPT_RT
# define __get_cpu_lock_lockdep_do(var, cpu)				\
	do {								\
		struct cpu_lock_t *__this_lock__ =			\
			&per_cpu(var##__lock, cpu);			\
		int __do_bh_disable__ = 0;				\
		if (!in_irq() && !irqs_disabled()) {			\
			local_bh_disable();				\
			__do_bh_disable__ = 1;				\
		}							\

# define __get_cpu_lock_lockdep_while()		\
		if (__do_bh_disable__)		\
			__this_lock__->bh = 1;	\
	} while (0)

# define __put_cpu_lock_lockdep_do(var, cpu)				\
	do {								\
		struct cpu_lock_t *__this_lock__ =			\
			&per_cpu(var##__lock, cpu);			\
		int __do_bh_enable__ = 0;				\
		if (__this_lock__->bh && !irqs_disabled()) {		\
			__do_bh_enable__ = 1;				\
			__this_lock__->bh = 0;				\
		}							\

# define __put_cpu_lock_lockdep_while()					\
		if (__do_bh_enable__)					\
			local_bh_enable();				\
	} while (0)
#else /* !PREEMPT_RT */
# define __get_cpu_lock_lockdep_do(var, cpu)	do {
# define __get_cpu_lock_lockdep_while()		} while (0)
# define __put_cpu_lock_lockdep_do(var, cpu)	do {
# define __put_cpu_lock_lockdep_while()		} while (0)
#endif

#define __get_cpu_lock_cpu(var, cpu, type)		\
	do {						\
		struct cpu_lock_t *__lock__ =		\
			&per_cpu(var##__lock, cpu);	\
		if (__lock__->owner == current) {	\
			static int once;		\
			__lock__->depth++;		\
			if (!once) {			\
				once++;			\
				printk("depth hit!\n");	\
				WARN_ON(1);		\
			}				\
			break;				\
		}					\
		acquire_lock##type(&__lock__->lock);	\
		__lock__->owner = current;		\
	} while (0)

#define get_cpu_lock_cpu(var, cpu)			\
	__get_cpu_lock_lockdep_do(var, cpu)		\
	__get_cpu_lock_cpu(var, cpu, );			\
	__get_cpu_lock_lockdep_while()

#define get_cpu_lock_cpu_irq(var, cpu)		\
	__get_cpu_lock_cpu(var, cpu, _irq)

#define get_cpu_lock_cpu_irqsave(var, cpu, flags)	\
	do {						\
		local_irq_save_nort(flags);		\
		get_cpu_lock_cpu(var, cpu);		\
	} while (0)

#define __get_cpu_lock(var, type)				\
	({							\
	   int __cpu;						\
	   do {							\
		   __cpu = raw_smp_processor_id();		\
		   get_cpu_lock_cpu##type(var, __cpu);		\
		   break;					\
	   } while (1);						\
	   __cpu;						\
	})

#define get_cpu_lock(var)			\
	__get_cpu_lock(var, )

#define get_cpu_lock_irq(var)			\
	__get_cpu_lock(var, _irq)

#define get_cpu_lock_irqsave(var, flags)	\
	({					\
		local_irq_save_nort(flags);	\
		get_cpu_lock(var);		\
	})

#define __put_cpu_lock(var, cpu, type)				\
	do {							\
		struct cpu_lock_t *__lock__ =			\
			&per_cpu(var##__lock, cpu);		\
		BUG_ON(__lock__->owner != current);		\
		if (__lock__->depth) {				\
			if (WARN_ON(__lock__->depth < 0))	\
				printk(KERN_WARNING		\
				       "lockdepth=%d\n",	\
				       __lock__->depth);	\
			__lock__->depth--;			\
			break;					\
		}						\
		__lock__->owner = NULL;				\
		release_lock##type(&__lock__->lock);		\
	} while (0)

#define put_cpu_lock(var, cpu)				\
	__put_cpu_lock_lockdep_do(var, cpu)		\
	__put_cpu_lock(var, cpu, );			\
	__put_cpu_lock_lockdep_while()


#define put_cpu_lock_irq(var, cpu)		\
	__put_cpu_lock(var, cpu, _irq)

#define put_cpu_lock_irqrestore(var, cpu, flags)		\
	do {							\
		put_cpu_lock(var, cpu);				\
		local_irq_restore_nort(flags);			\
	} while (0)

#define init_cpu_lock(var)					\
	do {							\
		int __cpu;					\
		for_each_possible_cpu(__cpu) {			\
			struct cpu_lock_t *lock =		\
				&per_cpu(var##__lock, __cpu);	\
			lock_init(&lock->lock);		\
			lock->owner = NULL;			\
			lock->depth = 0;			\
			lock->bh = 0;				\
		}						\
	} while (0)

#else /* !RT */

#define DEFINE_CPU_LOCK(lock)		int  __unused__##lock##__lock
#define DECLARE_CPU_LOCK(lock)		extern int  __unused__##lock##__lock
#define EXPORT_CPU_LOCK_SYMBOL(lock)
#define EXPORT_CPU_LOCK_SYMBOL_GPL(lock)

#define get_cpu_lock_cpu(var, cpu)			do { } while (0)
#define get_cpu_lock_cpu_irq(var, cpu)		\
	local_irq_disable()
#define get_cpu_lock_cpu_irqsave(var, cpu, flags)	\
	local_irq_save(flags)

#define get_cpu_lock(var)	get_cpu()
#define get_cpu_lock_irq(var)		 \
	({				 \
	  int __cpu;			 \
	  local_irq_disable();		 \
	  __cpu = get_cpu();		 \
	  __cpu;			 \
	})
#define get_cpu_lock_irqsave(var, flags) \
	({				 \
	  int __cpu;			 \
	  local_irq_save(flags);	 \
	  __cpu = get_cpu();		 \
	  __cpu;			 \
	})

#define put_cpu_lock(var, cpu)	put_cpu();
#define put_cpu_lock_irq(var, cpu)		 \
	do {					 \
		put_cpu();			 \
		local_irq_enable();		 \
	} while (0)
#define put_cpu_lock_irqrestore(var, cpu, flags) \
	do {					 \
		put_cpu();			 \
		local_irq_restore(flags);	 \
	} while (0)

#define init_cpu_lock(var)			\
	do {					\
		(void)__unused__##var##__lock;	\
	} while (0)
#endif /* RT */

#ifdef CONFIG_SMP

struct percpu_data {
	void *ptrs[1];
};

#define __percpu_disguise(pdata) (struct percpu_data *)~(unsigned long)(pdata)
/* 
 * Use this to get to a cpu's version of the per-cpu object dynamically
 * allocated. Non-atomic access to the current CPU's version should
 * probably be combined with get_cpu()/put_cpu().
 */ 
#define percpu_ptr(ptr, cpu)                              \
({                                                        \
        struct percpu_data *__p = __percpu_disguise(ptr); \
        (__typeof__(ptr))__p->ptrs[(cpu)];	          \
})

extern void *__percpu_alloc_mask(size_t size, gfp_t gfp, cpumask_t *mask);
extern void percpu_free(void *__pdata);

#else /* CONFIG_SMP */

#define percpu_ptr(ptr, cpu) ({ (void)(cpu); (ptr); })

static __always_inline void *__percpu_alloc_mask(size_t size, gfp_t gfp, cpumask_t *mask)
{
	return kzalloc(size, gfp);
}

static inline void percpu_free(void *__pdata)
{
	kfree(__pdata);
}

#endif /* CONFIG_SMP */

#define percpu_alloc_mask(size, gfp, mask) \
	__percpu_alloc_mask((size), (gfp), &(mask))

#define percpu_alloc(size, gfp) percpu_alloc_mask((size), (gfp), cpu_online_map)

/* (legacy) interface for use without CPU hotplug handling */

#define __alloc_percpu(size)	percpu_alloc_mask((size), GFP_KERNEL, \
						  cpu_possible_map)
#define alloc_percpu(type)	(type *)__alloc_percpu(sizeof(type))
#define free_percpu(ptr)	percpu_free((ptr))
#define per_cpu_ptr(ptr, cpu)	percpu_ptr((ptr), (cpu))

#endif /* __LINUX_PERCPU_H */
