#ifndef __LINUX_RT_LOCK_H
#define __LINUX_RT_LOCK_H

/*
 * Real-Time Preemption Support
 *
 * started by Ingo Molnar:
 *
 *  Copyright (C) 2004, 2005 Red Hat, Inc., Ingo Molnar <mingo@redhat.com>
 *
 * - Updated to work with lock_t
 *    Copyright (C) 2008 Red Hat, Inc., Steven Rostedt <srostedt@redhat.com>
 *
 * This file contains the main data structure definitions.
 */
#include <linux/rtmutex.h>
#include <asm/atomic.h>
#include <linux/spinlock_types.h>

#ifdef CONFIG_PREEMPT_RT
/*
 * rtlocks - an RT mutex plus lock-break field:
 */
typedef struct {
	struct rt_mutex		lock;
	unsigned int		break_lock;
#ifdef CONFIG_DEBUG_LOCK_ALLOC
	struct lockdep_map	dep_map;
#endif
} rtlock_t;

# define __RT_LOCK_UNLOCKED(name) \
	{ .lock = __RT_MUTEX_INITIALIZER((name).lock),	\
	  SPIN_DEP_MAP_INIT(name) }


#define RT_LOCK_UNLOCKED	__RT_LOCK_UNLOCKED(spin_old_style)

#define __DEFINE_RTLOCK(name) \
	rtlock_t name = __RT_LOCK_UNLOCKED(name)

#define DEFINE_RTLOCK(name) \
	rtlock_t name __cacheline_aligned_in_smp = __RT_LOCK_UNLOCKED(name)

/*
 * RW-semaphores are a spinlock plus a reader-depth count.
 *
 * Note that the semantics are different from the usual
 * Linux rw-sems, in PREEMPT_RT mode we do not allow
 * multiple readers to hold the lock at once, we only allow
 * a read-lock owner to read-lock recursively. This is
 * better for latency, makes the implementation inherently
 * fair and makes it simpler as well:
 */
struct rw_semaphore {
	struct rt_mutex		lock;
	int			read_depth;
#ifdef CONFIG_DEBUG_LOCK_ALLOC
	struct lockdep_map	dep_map;
#endif
};

/*
 * rwlocks - an RW semaphore plus lock-break field:
 */
typedef struct {
	struct rt_mutex		lock;
	int			read_depth;
	unsigned int		break_lock;
#ifdef CONFIG_DEBUG_LOCK_ALLOC
	struct lockdep_map	dep_map;
#endif
} rwlock_t;

#define __RW_LOCK_UNLOCKED(name) \
	{ .lock = __RT_MUTEX_INITIALIZER((name).lock),	\
	  RW_DEP_MAP_INIT(name) }

#define RW_LOCK_UNLOCKED	__RW_LOCK_UNLOCKED(rw_old_style)

#define DEFINE_RWLOCK(name) \
	rwlock_t name __cacheline_aligned_in_smp = __RW_LOCK_UNLOCKED(name)

/*
 * rwsems:
 */

#define __RWSEM_INITIALIZER(name) \
	{ .lock = __RT_MUTEX_INITIALIZER(name.lock), \
	  RW_DEP_MAP_INIT(name) }

#define DECLARE_RWSEM(lockname) \
	struct rw_semaphore lockname = __RWSEM_INITIALIZER(lockname)

extern void  __rt_rwsem_init(struct rw_semaphore *rwsem, char *name,
				     struct lock_class_key *key);

# define rt_init_rwsem(sem)				\
do {							\
	static struct lock_class_key __key;		\
							\
	__rt_rwsem_init((sem), #sem, &__key);		\
} while (0)

extern void  rt_down_write(struct rw_semaphore *rwsem);
extern void
rt_down_read_nested(struct rw_semaphore *rwsem, int subclass);
extern void
rt_down_write_nested(struct rw_semaphore *rwsem, int subclass);
extern void  rt_down_read(struct rw_semaphore *rwsem);
extern void  rt_down_read_non_owner(struct rw_semaphore *rwsem);
extern int  rt_down_write_trylock(struct rw_semaphore *rwsem);
extern int  rt_down_read_trylock(struct rw_semaphore *rwsem);
extern void  rt_up_read(struct rw_semaphore *rwsem);
extern void  rt_up_read_non_owner(struct rw_semaphore *rwsem);
extern void  rt_up_write(struct rw_semaphore *rwsem);
extern void  rt_downgrade_write(struct rw_semaphore *rwsem);

#define rt_rwsem_is_locked(rws)	(rt_mutex_is_locked(&(rws)->lock))
#define init_rwsem(rwsem) rt_init_rwsem(rwsem)

#define down_read(rwsem) rt_down_read(rwsem)
#define down_read_non_owner(rwsem) rt_down_read_non_owner(rwsem)
#define down_read_trylock(rwsem) rt_down_read_trylock(rwsem)
#define down_write(rwsem) rt_down_write(rwsem)
#define down_read_nested(rwsem, subclass) rt_down_read_nested(rwsem, subclass)
#define down_write_nested(rwsem, subclass) rt_down_write_nested(rwsem, subclass)
#define down_write_trylock(rwsem) rt_down_write_trylock(rwsem)

#define up_read(rwsem) rt_up_read(rwsem)
#define up_read_non_owner(rwsem) rt_up_read_non_owner(rwsem)
#define up_write(rwsem) rt_up_write(rwsem)

#define downgrade_write(rwsem) rt_downgrade_write(rwsem)
#define rwsem_is_locked(rwsem) rt_rwsem_is_locked(rwsem)

extern void __lockfunc rt_write_lock(rwlock_t *rwlock);
extern void __lockfunc rt_read_lock(rwlock_t *rwlock);
extern int __lockfunc rt_write_trylock(rwlock_t *rwlock);
extern int __lockfunc rt_read_trylock(rwlock_t *rwlock);
extern void __lockfunc rt_write_unlock(rwlock_t *rwlock);
extern void __lockfunc rt_read_unlock(rwlock_t *rwlock);
extern void
__rt_rwlock_init(rwlock_t *rwlock, char *name, struct lock_class_key *key);

#define _rwlock_init(rwl, n, f, l)			\
do {							\
	static struct lock_class_key __key;		\
							\
	__rt_rwlock_init(rwl, n, &__key);		\
} while (0)

#define rwlock_init(lock) _rwlock_init(lock, #lock, __FILE__, __LINE__)

#define write_lock(lock) 	rt_write_lock(lock)
#define write_lock_irqsave(lock, flags)				\
	do {							\
		(void)flags;					\
		rt_write_lock(lock);				\
	} while (0)
#define write_trylock(lock) 	__cond_lock(lock, rt_write_trylock(lock))
#define write_trylock_irqsave(lock, flags)			\
	({							\
		(void)flags;					\
		__cond_lock(lock, rt_write_trylock(lock));	\
	 })

#define read_lock(lock)		rt_read_lock(lock)
#define read_trylock(lock)	__cond_lock(lock, rt_read_trylock(lock))
#define read_lock_irqsave(lock, flags)		\
	({					\
		(void)flags;			\
		rt_read_lock(lock);		\
	 })

#define write_unlock(lock)		rt_write_unlock(lock)
#define write_unlock_irqrestore(lock, flags)			\
	({							\
		(void)flags;					\
		rt_write_unlock(lock);		\
	 })
#define read_unlock(lock)		rt_read_unlock(lock)
#define read_unlock_irqrestore(lock, flags)	\
	do {					\
		(void)flags;			\
		rt_read_unlock(lock);		\
	} while (0)

#define read_can_lock(l)		(!rt_mutex_is_locked(&(l)->lock))
#define write_can_lock(l)		((l)->lock.owner == NULL)

#define read_lock_irq(lock)		rt_read_lock(lock)
#define read_lock_bh(lock)		rt_read_lock(lock)

#define write_lock_irq(lock)		rt_write_lock(lock)
#define write_lock_bh(lock)		rt_write_lock(lock)

#define read_unlock_irq(lock)		rt_read_unlock(lock)
#define read_unlock_bh(lock)		rt_read_unlock(lock)

#define write_unlock_irq(lock)		rt_write_unlock(lock)
#define write_unlock_bh(lock)		rt_write_unlock(lock)

extern int cond_resched_rt_lock(rtlock_t *lock);
extern void rt_spin_lock(rtlock_t *lock);

#ifdef CONFIG_DEBUG_LOCK_ALLOC
extern void rt_spin_lock_nested(rtlock_t *lock, int subclass);
extern void _rt_spin_lock_nest_lock(rtlock_t *lock,
			    struct lockdep_map *nest_lock);
# define rt_spin_lock_nest_lock(lock, nest_lock)				\
	 do {								\
		 typecheck(struct lockdep_map *, &(nest_lock)->dep_map);\
		 _rt_spin_lock_nest_lock(lock, &(nest_lock)->dep_map);	\
	 } while (0)
#else
# define rt_spin_lock_nested(l, subclass)	rt_spin_lock(l)
# define rt_spin_lock_nest_lock(l, nest)	rt_spin_lock(l)
#endif

extern void rt_spin_unlock(rtlock_t *lock);
extern void rt_spin_unlock_wait(rtlock_t *lock);
extern int rt_spin_trylock(rtlock_t *lock);
extern void
__rt_spin_lock_init(rtlock_t *lock, char *name, struct lock_class_key *key);

#define _rt_spin_lock_init(sl, n, f, l) \
do {							\
	static struct lock_class_key __key;		\
							\
	__rt_spin_lock_init(sl, n, &__key);		\
} while (0)

#define rt_spin_lock_init(lock) _rt_spin_lock_init(lock, #lock, __FILE__, __LINE__)

#define rt_lock_is_locked(l)	(rt_mutex_is_locked(&(l)->lock))

extern int rt_atomic_dec_and_spin_lock(rtlock_t *lock, atomic_t *atomic);

/*
 * lockdep-less calls, for derived types like rwlock:
 * (for trylock they can use rt_mutex_trylock() directly.
 */
extern void __lockfunc __rt_spin_lock(struct rt_mutex *lock);
extern void __lockfunc __rt_spin_unlock(struct rt_mutex *lock);

#endif /* CONFIG_PREEMPT_RT */

#endif

